from abc import ABC, abstractmethod
from collections.abc import Sequence, Collection, Iterator
from typing import Any

from annet.generators import PartialGenerator
from annet.rpl import SingleCondition, MatchField, ThenField, RoutingPolicy, ConditionOperator
from .entities import (
    CommunityList, CommunityLogic, CommunityType, arista_well_known_community, mangle_united_community_list_name,
)


def get_used_community_lists(
        communities: Collection[CommunityList], policies: Collection[RoutingPolicy],
) -> list[CommunityList]:
    communities_dict = {c.name: c for c in communities}
    used_communities: set[str] = set()
    for policy in policies:
        for statement in policy.statements:
            condition: SingleCondition[Sequence[str]]
            for match_field in (
                    MatchField.community, MatchField.large_community,
                    MatchField.extcommunity_rt, MatchField.extcommunity_soo
            ):
                for condition in statement.match.find_all(match_field):
                    used_communities.update(condition.value)
            for then_field in (
                    ThenField.community, ThenField.large_community,
                    ThenField.extcommunity_rt, ThenField.extcommunity_soo,
                    ThenField.extcommunity,
            ):
                for action in statement.then.find_all(then_field):
                    if action.value.replaced is not None:
                        used_communities.update(action.value.replaced)
                    used_communities.update(action.value.added)
                    used_communities.update(action.value.removed)
    return [
        communities_dict[name] for name in sorted(used_communities)
    ]


def get_used_united_community_lists(
        communities: Collection[CommunityList], policies: Collection[RoutingPolicy],
) -> list[list[CommunityList]]:
    """
    Return communities united into groups according to HAS_ANY policy
    """
    communities_dict = {c.name: c for c in communities}
    used_communities: dict[str, list[CommunityList]] = {}
    for policy in policies:
        for statement in policy.statements:
            condition: SingleCondition[Sequence[str]]
            for match_field in (
                    MatchField.community, MatchField.large_community,
                    MatchField.extcommunity_rt, MatchField.extcommunity_soo
            ):
                for condition in statement.match.find_all(match_field):
                    if (
                            condition.operator == ConditionOperator.HAS_ANY and
                            len(condition.value) > 1
                    ):
                        united_name = mangle_united_community_list_name(condition.value)
                        united_communities: list[CommunityList] = [
                            communities_dict[name] for name in condition.value
                        ]
                        if not all(united_communities[0].type == c.type for c in united_communities):
                            raise ValueError(
                                f"Cannot apply HAS_ANY to communities of different types, "
                                f"found for policy: `{policy.name}`, statement: {statement.name}"
                            )
                        if not all(united_communities[0].use_regex == c.use_regex for c in united_communities):
                            raise ValueError(
                                f"Cannot apply HAS_ANY to communities with different use_regex flag, "
                                f"found for policy: `{policy.name}`, statement: {statement.name}"
                            )
                        used_communities[united_name] = united_communities
                    else:
                        for name in condition.value:
                            used_communities[name] = [communities_dict[name]]
            for then_field in (
                    ThenField.community, ThenField.large_community,
                    ThenField.extcommunity_rt, ThenField.extcommunity_soo
            ):
                for action in statement.then.find_all(then_field):
                    if action.value.replaced is not None:
                        for name in action.value.replaced:
                            used_communities[name] = [communities_dict[name]]
                    for name in action.value.added:
                        used_communities[name] = [communities_dict[name]]
                    for name in action.value.removed:
                        used_communities[name] = [communities_dict[name]]
    return [
        used_communities[name] for name in sorted(used_communities)
    ]


class CommunityListGenerator(PartialGenerator, ABC):
    TAGS = ["policy", "rpl", "routing"]

    @abstractmethod
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        raise NotImplementedError()

    @abstractmethod
    def get_community_lists(self, device: Any) -> list[CommunityList]:
        raise NotImplementedError()

    def get_used_community_lists(self, device: Any) -> list[CommunityList]:
        return get_used_community_lists(
            communities=self.get_community_lists(device),
            policies=self.get_policies(device),
        )

    def get_used_united_community_lists(self, device: Any) -> list[list[CommunityList]]:
        return get_used_united_community_lists(
            communities=self.get_community_lists(device),
            policies=self.get_policies(device),
        )

    def acl_huawei(self, _):
        return r"""
        ip community-filter
        ip extcommunity-filter
        ip extcommunity-list
        ip large-community-filter
        """

    def ref_huawei(self, _):
        return """
        route-policy
            if-match community-filter <name>
            if-match extcommunity-filter <name>
            if-match extcommunity-list soo <name>
            if-match large-community-filter <name>
            apply comm-filter <name>
        """

    def _huawei_community_filter(self, index: int, community_list: CommunityList, members: str) -> Sequence[str]:
        if community_list.use_regex:
            match_type = "advanced"
        else:
            match_type = "basic"
        if community_list.type is CommunityType.BASIC:
            return "ip community-filter", match_type, community_list.name, f"index {index}", "permit", members
        elif community_list.type is CommunityType.RT:
            return "ip extcommunity-filter", match_type, community_list.name, f"index {index}", "permit", members
        elif community_list.type is CommunityType.SOO:
            return "ip extcommunity-list soo", match_type, community_list.name, f"index {index}", "permit", members
        elif community_list.type is CommunityType.LARGE:
            return "ip large-community-filter", match_type, community_list.name, f"index {index}", "permit", members
        else:
            raise NotImplementedError(f"CommunityList type {community_list.type} not implemented for huawei")

    def run_huawei(self, device: Any):
        for community_list in self.get_used_community_lists(device):
            if community_list.use_regex and len(community_list.members) > 1:
                raise NotImplementedError("Multiple regex is not supported for huawei")
            if community_list.type is CommunityType.RT:
                # RT communities used with prefix rt
                members: Sequence[str] = [f"rt {m}" for m in community_list.members]
            else:
                members = community_list.members

            if community_list.logic == CommunityLogic.AND:
                # to get AND logic the communities should be in one sting
                yield self._huawei_community_filter(10, community_list, " ".join(members))
            elif community_list.logic == CommunityLogic.OR:
                for i, member in enumerate(members):
                    member_id = (i + 1) * 10
                    yield self._huawei_community_filter(member_id, community_list, member)
            else:
                raise NotImplementedError(f"Community logic {community_list.logic} is not implemented for huawei")

    def acl_arista(self, _):
        return r"""
        ip community-list
        ip extcommunity-list
        """

    def _arista_community_list(
            self, name: str, use_regex: bool, comm_type: CommunityType, members: str,
    ) -> Sequence[str]:
        if use_regex:
            match_type = "regexp"
        else:
            match_type = ""
        if comm_type is CommunityType.BASIC:
            return "ip community-list", match_type, name, "permit", members
        elif comm_type is CommunityType.RT:
            return "ip extcommunity-list", match_type, name, "permit", members
        elif comm_type is CommunityType.SOO:
            return "ip extcommunity-list", match_type, name, "permit", members
        elif comm_type is CommunityType.LARGE:
            return "ip large-community-list", match_type, name, "permit", members
        else:
            raise NotImplementedError(f"CommunityList type {comm_type} not implemented for arista")

    def _arista_community_prefix(self, community_list: CommunityList) -> str:
        if community_list.type is CommunityType.BASIC:
            return ""
        elif community_list.type is CommunityType.RT:
            if community_list.use_regex:
                return "RT:"
            return "rt "
        elif community_list.type is CommunityType.SOO:
            if community_list.use_regex:
                return "SoO:"
            return "soo "
        elif community_list.type is CommunityType.LARGE:
            return ""
        else:
            raise NotImplementedError(f"CommunityList type {community_list.type} not implemented for arista")

    def run_arista(self, device):
        for community_list_union in self.get_used_united_community_lists(device):
            name = mangle_united_community_list_name([c.name for c in community_list_union])
            for community_list in community_list_union:
                if community_list.use_regex and len(community_list.members) > 1:
                    raise NotImplementedError("Multiple regex is not supported for arista")

                member_prefix = self._arista_community_prefix(community_list)

                if community_list.logic == CommunityLogic.AND:
                    # to get AND logic the communities should be in one sting
                    member_str = " ".join(
                        member_prefix + arista_well_known_community(m)
                        for m in community_list.members
                    )
                    yield self._arista_community_list(
                        name=name,
                        use_regex=community_list.use_regex,
                        comm_type=community_list.type,
                        members=member_str,
                    )
                elif community_list.logic == CommunityLogic.OR:
                    for member in community_list.members:
                        yield self._arista_community_list(
                            name=name,
                            use_regex=community_list.use_regex,
                            comm_type=community_list.type,
                            members=member_prefix + member,
                        )
                else:
                    raise NotImplementedError(f"Community logic {community_list.logic} is not implemented for arista")

    def acl_iosxr(self, _) -> str:
        return r"""
        community-set *
            ~ %global=1
        extcommunity-set *
            ~ %global=1
        """

    def _iosxr_community_list(self, community_list: CommunityList) -> Iterator[Sequence[str]]:
        if community_list.type is CommunityType.BASIC:
            name = "community-set"
        elif community_list.type is CommunityType.RT:
            name = "extcommunity-set rt"
        elif community_list.type is CommunityType.SOO:
            name = "extcommunity-set soo"
        else:
            raise NotImplementedError(f"CommunityList type {community_list.type} not implemented for Cisco IOS XR")

        with self.block(name, community_list.name):
            for n, community in enumerate(community_list.members):
                if n + 1 < len(community_list.members):
                    comma = ","
                else:
                    comma = ""
                if community_list.use_regex:
                    yield "ios-regex", f"'{community}'"
                else:
                    yield f"{community}{comma}",

    def run_iosxr(self, device):
        for community_list in self.get_used_community_lists(device):
            yield from self._iosxr_community_list(community_list)



from annet.annlib.command import Command, CommandList
from annet.annlib.netdev.views.hardware import HardwareView
from annet.vendors.tabparser import ArubaFormatter
from annet.vendors.base import AbstractVendor
from annet.vendors.registry import registry


@registry.register
class ArubaVendor(AbstractVendor):
    NAME = "aruba"

    def apply(self, hw: HardwareView, do_commit: bool, do_finalize: bool, path: str) -> tuple[CommandList, CommandList]:
        before, after = CommandList(), CommandList()

        before.add_cmd(Command("conf t"))
        after.add_cmd(Command("end"))
        if do_commit:
            after.add_cmd(Command("commit apply"))
        if do_finalize:
            after.add_cmd(Command("write memory"))

        return before, after

    def match(self) -> list[str]:
        return ["Aruba"]

    @property
    def reverse(self) -> str:
        return "no"

    @property
    def hardware(self) -> HardwareView:
        return HardwareView("Aruba")

    def make_formatter(self, **kwargs) -> ArubaFormatter:
        return ArubaFormatter(**kwargs)

    @property
    def exit(self) -> str:
        return "exit"

    def diff(self, order: bool) -> str:
        return "common.ordered_diff" if order else "aruba.default_diff"



import ssl
from typing import Optional

from adaptix import P
from adaptix.conversion import get_converter, link, link_constant, link_function
from annetbox.v42 import client_sync
from annetbox.v42 import models as api_models

from annet.adapters.netbox.common.adapter import NetboxAdapter, get_device_breed, get_device_hw
from annet.adapters.netbox.common.storage_base import BaseNetboxStorage
from annet.adapters.netbox.common.storage_opts import NetboxStorageOpts
from annet.adapters.netbox.v41.models import FHRPGroupAssignmentV41, FHRPGroupV41
from annet.adapters.netbox.v42.models import InterfaceV42, NetboxDeviceV42, PrefixV42, IpAddressV42, Vlan, Vrf
from annet.storage import Storage


class NetboxV42Adapter(NetboxAdapter[
    NetboxDeviceV42, InterfaceV42, IpAddressV42, PrefixV42,
    FHRPGroupV41, FHRPGroupAssignmentV41,
]):
    def __init__(
            self,
            storage: Storage,
            url: str,
            token: str,
            ssl_context: ssl.SSLContext | None,
            threads: int,
    ):
        self.netbox = client_sync.NetboxV42(url=url, token=token, ssl_context=ssl_context, threads=threads)
        self.convert_device = get_converter(
            api_models.Device,
            NetboxDeviceV42,
            recipe=[
                link_function(get_device_breed, P[NetboxDeviceV42].breed),
                link_function(get_device_hw, P[NetboxDeviceV42].hw),
                link_constant(P[NetboxDeviceV42].interfaces, factory=list),
                link_constant(P[NetboxDeviceV42].storage, value=storage),
                link(P[api_models.Device].name, P[NetboxDeviceV42].hostname),
                link(P[api_models.Device].name, P[NetboxDeviceV42].fqdn),
            ]
        )
        self.convert_interfaces = get_converter(
            list[api_models.Interface],
            list[InterfaceV42],
            recipe=[
                link_constant(P[InterfaceV42].ip_addresses, factory=list),
                link_constant(P[InterfaceV42].fhrp_groups, factory=list),
                link_constant(P[InterfaceV42].lag_min_links, value=None),
            ]
        )
        self.convert_ip_addresses = get_converter(
            list[api_models.IpAddress],
            list[IpAddressV42],
            recipe=[
                link_constant(P[IpAddressV42].prefix, value=None),
            ]
        )
        self.convert_ip_prefixes = get_converter(
            list[api_models.Prefix],
            list[PrefixV42],
        )
        self.convert_fhrp_group_assignments = get_converter(
            list[api_models.FHRPGroupAssignmentBrief],
            list[FHRPGroupAssignmentV41],
            recipe=[
                link_constant(P[FHRPGroupAssignmentV41].group, value=None),
                link_function(lambda model: model.group.id, P[FHRPGroupAssignmentV41].fhrp_group_id),
            ]
        )
        self.convert_fhrp_groups = get_converter(
            list[api_models.FHRPGroup],
            list[FHRPGroupV41],
        )

    def list_all_fqdns(self) -> list[str]:
        return [
            d.name
            for d in self.netbox.dcim_all_devices_brief().results
        ]

    def list_devices(self, query: dict[str, list[str]]) -> list[NetboxDeviceV42]:
        return [
            self.convert_device(dev)
            for dev in self.netbox.dcim_all_devices(**query).results
        ]

    def get_device(self, device_id: int) -> NetboxDeviceV42:
        return self.convert_device(self.netbox.dcim_device(device_id))

    def list_interfaces_by_devices(self, device_ids: list[int]) -> list[InterfaceV42]:
        return self.convert_interfaces(self.netbox.dcim_all_interfaces(device_id=device_ids).results)

    def list_interfaces(self, ids: list[int]) -> list[InterfaceV42]:
        return self.convert_interfaces(self.netbox.dcim_all_interfaces(id=ids).results)

    def list_ipaddr_by_ifaces(self, iface_ids: list[int]) -> list[IpAddressV42]:
        return self.convert_ip_addresses(self.netbox.ipam_all_ip_addresses(interface_id=iface_ids).results)

    def list_ipprefixes(self, prefixes: list[str]) -> list[PrefixV42]:
        return self.convert_ip_prefixes(self.netbox.ipam_all_prefixes(prefix=prefixes).results)

    def list_all_vrfs(self) -> list[Vrf]:
        return self.netbox.ipam_all_vrfs().results

    def list_all_vlans(self) -> list[Vlan]:
        return self.netbox.ipam_all_vlans().results

    def list_fhrp_group_assignments(
            self, iface_ids: list[int],
    ) -> list[FHRPGroupAssignmentV41]:
        raw_assignments = self.netbox.ipam_all_fhrp_group_assignments_by_interface(
            interface_id=iface_ids,
        )
        return self.convert_fhrp_group_assignments(raw_assignments.results)

    def list_fhrp_groups(self, ids: list[int]) -> list[FHRPGroupV41]:
        raw_groups = self.netbox.ipam_all_fhrp_groups_by_id(id=list(ids))
        return self.convert_fhrp_groups(raw_groups.results)


class NetboxStorageV42(BaseNetboxStorage[
    NetboxDeviceV42, InterfaceV42, IpAddressV42, PrefixV42,
    FHRPGroupV41, FHRPGroupAssignmentV41,
]):
    netbox: NetboxV42Adapter

    def __init__(self, opts: Optional[NetboxStorageOpts] = None):
        super().__init__(opts)
        self._all_vlans: list[Vlan] | None = None
        self._all_vrfs: list[Vrf] | None = None

    def _init_adapter(
            self,
            url: str,
            token: str,
            ssl_context: ssl.SSLContext | None,
            threads: int,
    ) -> NetboxAdapter[
        NetboxDeviceV42, InterfaceV42, IpAddressV42, PrefixV42,
        FHRPGroupV41, FHRPGroupAssignmentV41,
    ]:
        return NetboxV42Adapter(self, url, token, ssl_context, threads)

    def resolve_all_vlans(self) -> list[Vlan]:
        if self._all_vlans is None:
            self._all_vlans = self.netbox.list_all_vlans()
        return self._all_vlans

    def resolve_all_vrfs(self) -> list[Vrf]:
        if self._all_vrfs is None:
            self._all_vrfs = self.netbox.list_all_vrfs()
        return self._all_vrfs



from abc import ABC, abstractmethod
from collections.abc import Sequence, Iterable
from typing import Any, Literal

from annet.generators import PartialGenerator
from annet.rpl import PrefixMatchValue, MatchField, SingleCondition, RoutingPolicy
from .entities import IpPrefixList, PrefixListNameGenerator


class PrefixListFilterGenerator(PartialGenerator, ABC):
    TAGS = ["policy", "rpl", "routing"]

    @abstractmethod
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        raise NotImplementedError()

    @abstractmethod
    def get_prefix_lists(self, device: Any) -> Sequence[IpPrefixList]:
        raise NotImplementedError()

    # huawei
    def acl_huawei(self, _):
        return r"""
        ip ip-prefix
        ip ipv6-prefix
        """

    def _huawei_prefix_list(
            self,
            prefix_type: Literal["ipv6-prefix", "ip-prefix"],
            plist: IpPrefixList,
    ) -> Iterable[Sequence[str]]:
        for i, m in enumerate(plist.members):
            ge, le = m.or_longer
            yield (
                "ip",
                prefix_type,
                plist.name,
                f"index {i * 5 + 5}",
                "permit",
                str(m.prefix.network_address).upper(),
                str(m.prefix.prefixlen),
            ) + (
                ("greater-equal", str(ge)) if ge is not None else ()
            ) + (
                ("less-equal", str(le)) if le is not None else ()
            )

    def run_huawei(self, device: Any):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)

        name_generator = PrefixListNameGenerator(prefix_lists, policies)
        processed_names = set()
        for policy in policies:
            for statement in policy.statements:
                cond: SingleCondition[PrefixMatchValue]
                for cond in statement.match.find_all(MatchField.ip_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._huawei_prefix_list("ip-prefix", plist)
                        processed_names.add(plist.name)
                for cond in statement.match.find_all(MatchField.ipv6_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._huawei_prefix_list("ipv6-prefix", plist)
                        processed_names.add(plist.name)

    # arista
    def acl_arista(self, _):
        return r"""
        ip prefix-list
            seq
        ipv6 prefix-list
            seq
        """

    def _arista_prefix_list(
            self,
            prefix_type: Literal["ipv6", "ip"],
            plist: IpPrefixList,
    ) -> Iterable[Sequence[str]]:
        with self.block(prefix_type, "prefix-list", plist.name):
            for i, m in enumerate(plist.members):
                ge, le = m.or_longer
                yield (
                    f"seq {i * 10 + 10}",
                    "permit",
                    str(m.prefix),
                ) + (
                    ("ge", str(ge)) if ge is not None else ()
                ) + (
                    ("le", str(le)) if le is not None else ()
                )

    def run_arista(self, device: Any):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)
        name_generator = PrefixListNameGenerator(prefix_lists, policies)
        processed_names = set()
        for policy in policies:
            for statement in policy.statements:
                cond: SingleCondition[PrefixMatchValue]
                for cond in statement.match.find_all(MatchField.ip_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._arista_prefix_list("ip", plist)
                        processed_names.add(plist.name)
                for cond in statement.match.find_all(MatchField.ipv6_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._arista_prefix_list("ipv6", plist)
                        processed_names.add(plist.name)

    # Cisco IOS XR
    def acl_iosxr(self, device: Any):
        return r"""
        prefix-set
            ~ %global=1
        """

    def _iosxr_prefixlist(self, prefixlist: IpPrefixList):
        with self.block("prefix-set", prefixlist.name):
            for n, member in enumerate(prefixlist.members):
                if n + 1 < len(prefixlist.members):
                    comma = ","
                else:
                    comma = ""

                ge, le = member.or_longer
                if ge is le is None:
                    yield f"{member.prefix}{comma}",
                elif ge is None:
                    yield f"{member.prefix} le {le}{comma}",
                elif le is None:
                    yield f"{member.prefix} ge {ge}{comma}",
                elif ge == le:
                    yield f"{member.prefix} eq {ge}{comma}",
                else:
                    yield f"{member.prefix} ge {ge} le {le}{comma}",

    def run_iosxr(self, device: Any):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)

        name_generator = PrefixListNameGenerator(prefix_lists, policies)
        processed_names = set()
        for policy in policies:
            for statement in policy.statements:
                cond: SingleCondition[PrefixMatchValue]
                for cond in statement.match.find_all(MatchField.ip_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._iosxr_prefixlist(plist)
                        processed_names.add(plist.name)
                for cond in statement.match.find_all(MatchField.ipv6_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._iosxr_prefixlist(plist)
                        processed_names.add(plist.name)



import asyncio
import os
import statistics
from abc import ABC, abstractmethod
from functools import partial
from operator import itemgetter
from typing import Any, List, Optional

import colorama
from annet.annlib.command import Command, CommandList, Question  # noqa: F401


class CommandResult(ABC):
    @abstractmethod
    def get_out(self) -> str:
        pass


class ExecutorException(Exception):
    def __init__(self, *args: List[Any], auxiliary: Optional[Any] = None, **kwargs: object):
        self.auxiliary = auxiliary
        super().__init__(*args, **kwargs)

    def __repr__(self) -> str:
        return "%s(args=%r,auxiliary=%s)" % (self.__class__.__name__, self.args, self.auxiliary)


class ExecException(ExecutorException):
    def __init__(self, msg: str, cmd: str, res: str, **kwargs):
        super().__init__(**kwargs)
        self.args = msg, cmd, res
        self.kwargs = kwargs
        self.msg = msg
        self.cmd = cmd
        self.res = res

    def __str__(self) -> str:
        return str(self.msg)

    def __repr__(self) -> str:
        return "%s<%s, %s>" % (self.__class__.__name__, self.msg, self.cmd)


class BadCommand(ExecException):
    pass


class NonzeroRetcode(ExecException):
    pass


class CommitException(ExecException):
    pass


def _show_type_summary(caption, items, total, stat_items=None):
    if items:
        if not stat_items:
            stat = ""
        else:
            avg = statistics.mean(stat_items)
            stat = "   %(min).1f/%(max).1f/%(avg).1f/%(stdev)s  (min/max/avg/stdev)" % dict(
                min=min(stat_items),
                max=max(stat_items),
                avg=avg,
                stdev="-" if len(stat_items) < 2 else "%.1f" % statistics.stdev(stat_items, xbar=avg)
            )

        print("%-8s %d of %d%s" % (caption, len(items), total, stat))


def show_bulk_report(hostnames, res, durations, log_dir):
    total = len(hostnames)
    if not total:
        return

    colorama.init()

    print("\n====== bulk deploy report ======")

    done = [host for (host, hres) in res.items() if not isinstance(hres, Exception)]
    cancelled = [host for (host, hres) in res.items() if isinstance(hres, asyncio.CancelledError)]
    failed = [host for (host, hres) in res.items() if isinstance(hres, Exception) and host not in cancelled]
    lost = [host for host in hostnames if host not in res]
    limit = 30

    _show_type_summary("Done :", done, total, [durations[h] for h in done])
    _print_limit(done, partial(_print_hostname, style=colorama.Fore.GREEN), limit, total)

    _show_type_summary("Failed :", failed, total, [durations[h] for h in failed])

    _print_limit(failed, partial(_print_failed, res=res), limit, total)

    _show_type_summary("Cancelled :", cancelled, total, [durations[h] for h in cancelled if durations[h] is not None])
    _print_limit(cancelled, partial(_print_hostname, style=colorama.Fore.RED), limit, total)

    _show_type_summary("Lost :", lost, total)
    _print_limit(lost, _print_hostname, limit, total)

    err_limit = 5
    if failed:
        errs = {}
        for hostname in failed:
            fmt_err = _format_exc(res[hostname])
            if fmt_err in errs:
                errs[fmt_err] += 1
            else:
                errs[fmt_err] = 1
        print("Top errors :")
        for fmt_err, n in sorted(errs.items(), key=itemgetter(1), reverse=True)[:err_limit]:
            print("  %-4d %s" % (n, fmt_err))
        print("\n", end="")

    if log_dir:
        print("See deploy logs in %s/\n" % os.path.relpath(log_dir))


def _format_exc(exc):
    if isinstance(exc, ExecException):
        cmd = str(exc.cmd)
        if len(cmd) > 50:
            cmd = cmd[:50] + "~.."
        return "'%s', cmd '%s'" % (exc.msg, cmd)
    elif isinstance(exc, ExecutorException):
        return "%s%r" % (exc.__class__.__name__, exc.args)  # исключить многословный auxiliary
    else:
        return repr(exc)


def _print_hostname(host, style=None):
    if style:
        host = style + host + colorama.Style.RESET_ALL
    print("  %s" % host)


def _print_limit(items, printer, limit, total, end="\n"):
    if not items:
        return
    if len(items) > limit and len(items) > total * 0.7:
        print("  ... %d hosts" % len(items))
    for host in items[:limit]:
        printer(host)
    if len(items) > limit:
        print("  ... %d more hosts" % (len(items) - limit))

    print(end, end="")


def _print_failed(host, res):
    exc = res[host]
    color = colorama.Fore.YELLOW if isinstance(exc, Warning) else colorama.Fore.RED
    print("  %s - %s" % (color + host + colorama.Style.RESET_ALL, _format_exc(exc)))


class DeferredFileWrite:
    def __init__(self, file, mode="r"):
        self._file = file
        wrapper = {"w": "a", "wb": "ab"}
        if mode in wrapper:
            self._mode = wrapper[mode]
        else:
            raise Exception()

    def write(self, data):
        with open(self._file, self._mode, encoding="utf-8") as fh:
            fh.write(data)

    def close(self):
        pass

    def flush(self):
        pass



from ipaddress import IPv4Network, IPv6Network, ip_network
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass
from enum import Enum
from typing import Optional, List

from annet.rpl import RoutingPolicy, PrefixMatchValue, OrLonger


class CommunityLogic(Enum):
    AND = "AND"
    OR = "OR"


class CommunityType(Enum):
    BASIC = "BASIC"
    RT = "RT"
    SOO = "SOO"
    COST = "COST"
    LARGE = "LARGE"


@dataclass(frozen=True)
class CommunityList:
    name: str
    members: Sequence[str]
    type: CommunityType = CommunityType.BASIC
    logic: CommunityLogic = CommunityLogic.OR
    use_regex: bool = False


@dataclass(frozen=True)
class RDFilter:
    name: str
    number: int
    members: Sequence[str]


@dataclass(frozen=True)
class AsPathFilter:
    name: str
    filters: Sequence[str]


@dataclass
class IpPrefixListMember:
    prefix: IPv4Network | IPv6Network
    or_longer: OrLonger = (None, None)

    def __post_init__(self):
        self.prefix = ip_network(self.prefix)


@dataclass
class IpPrefixList:
    name: str
    members: list[IpPrefixListMember]

    def __post_init__(self):
        for (i, m) in enumerate(self.members):
            if isinstance(m, str):
                self.members[i] = IpPrefixListMember(m)


def ip_prefix_list(
        name: str,
        members_or_str: Sequence[IpPrefixListMember | str],
        or_longer: OrLonger = (None, None),
) -> IpPrefixList:
    members: List[IpPrefixListMember] = []
    for m in members_or_str:
        if isinstance(m, str):
            m = IpPrefixListMember(
                prefix=ip_network(m),
                or_longer=or_longer,
            )
        elif m.or_longer == (None, None):
            m = IpPrefixListMember(
                prefix=m.prefix,
                or_longer=or_longer,
            )
        members.append(m)
    return IpPrefixList(
        name=name,
        members=members,
    )


def arista_well_known_community(community: str) -> str:
    if community == "65535:0":
        return "GSHUT"
    return community


def mangle_united_community_list_name(values: Sequence[str]) -> str:
    """Name for a list used as HAS_ANY between multiple lists"""
    return "_OR_".join(values)


class PrefixListNameGenerator:
    def __init__(self, prefix_lists: Sequence[IpPrefixList], policies: Sequence[RoutingPolicy]):
        self._prefix_lists = {x.name: x for x in prefix_lists}
        self._policies = {x.name: x for x in policies}  # this is here for a later use ~azryve@

    def get_prefix(self, name: str, match: PrefixMatchValue) -> IpPrefixList:
        orig_prefix = self._prefix_lists[name]
        override_name: Optional[str] = None
        override_orlonger: Optional[OrLonger] = None

        if any(match.or_longer):
            ge, le = match.or_longer
            ge_str = "unset" if ge is None else str(ge)
            le_str = "unset" if le is None else str(le)
            override_name = f"{orig_prefix.name}_{ge_str}_{le_str}"
            override_orlonger = match.or_longer

        return IpPrefixList(
            name=override_name or name,
            members=[
                IpPrefixListMember(
                    x.prefix,
                    or_longer=override_orlonger or x.or_longer,
                )
                for x in orig_prefix.members
            ],
        )


def group_community_members(
    all_communities: dict[str, CommunityList], communities: list[str],
) -> dict[CommunityType, list[str]]:
    members: dict[CommunityType, list[str]] = defaultdict(list)
    for community_name in communities:
        community = all_communities[community_name]
        members[community.type].extend(community.members)
    return members
